""" Generators for basis evaluations """

from dune.perftool.generation import (cached,
                                      class_member,
                                      domain,
                                      generator_factory,
                                      iname,
                                      include_file,
                                      instruction,
                                      preamble,
                                      symbol,
                                      temporary_variable,
                                      )
from dune.perftool.pdelab.quadrature import (name_quadrature_position_in_cell,
                                             quadrature_iname,
                                             )
from dune.perftool.pdelab.geometry import (dimension_iname,
                                           name_dimension,
                                           name_jacobian_inverse_transposed,
                                           to_cell_coordinates,
                                           )
from dune.perftool.pdelab.localoperator import (lop_template_ansatz_gfs,
                                                lop_template_test_gfs,
                                                )
from dune.perftool.pdelab.driver import FEM_name_mangling
from dune.perftool.pdelab import restricted_name
from pymbolic.primitives import Product, Subscript, Variable
from loopy import Reduction


@symbol
def type_localbasis_cache(element):
    return "LocalBasisCacheWithoutReferences<{}>".format(type_gfs(element))


@class_member("operator")
def define_localbasis_cache(element, name):
    include_file("dune/perftool/localbasiscache.hh", filetag="operatorfile")
    t = type_localbasis_cache(element)
    return "{} {};".format(t, name)


@symbol
def name_localbasis_cache(element):
    name = "cache_{}".format(FEM_name_mangling(element))
    define_localbasis_cache(element, name)
    return name


@preamble
def define_lfs_bound(lfs, bound):
    return 'auto {} = {}.size();'.format(bound, lfs)


@symbol
def name_lfs_bound(lfs):
    # LFS might either be given by an UFL element or by a string describig its name
    from ufl import FiniteElementBase
    if isinstance(lfs, FiniteElementBase):
        return name_lfs_bound(name_lfs(lfs))

    bound = '{}_size'.format(lfs)
    define_lfs_bound(lfs, bound)
    return bound


@generator_factory(cache_key_generator=lambda e, r, **kw: (e, r))
def name_lfs(element, restriction, prefix=None):
    # Omitting the prefix is only valid upon a second call, which will
    # result in a cache hit.
    assert prefix

    # Additionally, element is expected to be a ufl finite element
    from ufl import FiniteElementBase
    assert isinstance(element, FiniteElementBase)

    # Recurse into the given element to define all other local function spaces!
    from ufl import MixedElement
    if isinstance(element, MixedElement):
        for i, subelem in enumerate(element.sub_elements()):
            # TODO in this case, we need to trigger the extraction mechanism
            # as preambles in our code.
            name_lfs(subelem, restriction, prefix + "_" + str(i))

    # Now trigger the creation of all those other symbols/preables necessary for this lfs

    # Now return the prefix!
    return prefix


@generator_factory(cache_key_generator=lambda e, **kw: e)
def type_gfs(element, basetype=None, index_stack=None):
    # Omitting basetype and index_stack is only valid upon a second call,
    # which will result in a cache hit.
    assert basetype
    assert index_stack is not None

    # Additionally, element is expected to be a ufl finite element
    from ufl import FiniteElementBase
    assert isinstance(element, FiniteElementBase)

    # Recurse into the given element to define all other local function spaces!
    from ufl import MixedElement
    if isinstance(element, MixedElement):
        for i, subelem in enumerate(element.sub_elements()):
            type_gfs(subelem, basetype=basetype, index_stack=index_stack + (i,))

    if len(index_stack) == 0:
        return basetype
    else:
        include_file("dune/typetree/childextraction.hh", filetag="operatorfile")
        return 'Dune::TypeTree::Child<{},{}>'.format(basetype, ','.join(str(i) for i in index_stack))


def traverse_lfs_tree(arg):
    from dune.perftool.ufl.modified_terminals import ModifiedArgumentDescriptor
    assert isinstance(arg, ModifiedArgumentDescriptor)

    # First we need to determine the basename as given in the signature of
    # this kernel method!
    from dune.perftool.pdelab.argument import name_argumentspace
    lfs_basename = name_argumentspace(arg)
    from dune.perftool.pdelab.localoperator import lop_template_gfs
    gfs_basename = lop_template_gfs(arg)[1]

    # Now start recursively extracting local function spaces and fill the cache with
    # all those values. That way we can later get a correct local function space with
    # just the ufl finite element.
    name_lfs(arg.argexpr.element(), arg.restriction, prefix=lfs_basename)
    type_gfs(arg.argexpr.element(), basetype=gfs_basename, index_stack=())


@generator_factory(item_tags=("iname",), cache_key_generator=lambda e, r, c: (e, c))
def _lfs_iname(element, restriction, context):
    lfs = name_lfs(element, restriction)
    bound = name_lfs_bound(lfs)

    name = "{}_{}_index".format(lfs, context)
    domain(name, bound)

    return name


def lfs_iname(element, restriction, count=None, context=''):
    """ Get the iname to iterate over the local function space given by element

    Arguments:
    ----------
    element: ufl.FiniteElementBase
        The finite element this local function space belongs to
    argcount: int
        Use to realize double nesting in case of jacobians
    context: str
        Some generation methods will require you to duplicate an iname for
        a given purpose, see the 'Loops and dependencies' of the loopy docs:
        https://documen.tician.de/loopy/tutorial.html#loops-and-dependencies
    """
    assert not ((context == '') and (count is None))
    if count is not None:
        if context != '':
            context = "{}_{}".format(count, context)
        else:
            context = str(count)
    return _lfs_iname(element, restriction, context)


@preamble
def declare_cache_temporary(element, restriction, name, which):
    t_cache = type_localbasis_cache(element)
    lfs = name_lfs(element, restriction)
    return "typename {}::{}ReturnType {};".format(t_cache,
                                                  which,
                                                  name,
                                                  )


@cached
def evaluate_basis(element, name, restriction):
    lfs = name_lfs(element, restriction)
    temporary_variable(name, shape=(name_lfs_bound(lfs),), decl_method=None)
    declare_cache_temporary(element, restriction, name, which='Function')
    cache = name_localbasis_cache(element)
    qp = name_quadrature_position_in_cell(restriction)
    instruction(inames=(quadrature_iname(),
                        ),
                code='{} = {}.evaluateFunction({}, {}.finiteElement().localBasis());'.format(name,
                                                                                             cache,
                                                                                             qp,
                                                                                             lfs,
                                                                                             ),
                assignees=frozenset({name}),
                read_variables=frozenset({qp}),
                )


@symbol
def name_basis(element, restriction):
    name = restricted_name("phi", restriction)
    evaluate_basis(element, name, restriction)
    return name


@cached
def evaluate_reference_gradient(element, name, restriction):
    lfs = name_lfs(element, restriction)
    temporary_variable(name, shape=(name_lfs_bound(lfs), name_dimension()), decl_method=None)
    declare_cache_temporary(element, restriction, name, which='Jacobian')
    cache = name_localbasis_cache(element)
    qp = name_quadrature_position_in_cell(restriction)
    instruction(inames=(quadrature_iname(),
                        ),
                code='{} = {}.evaluateJacobian({}, {}.finiteElement().localBasis());'.format(name,
                                                                                             cache,
                                                                                             qp,
                                                                                             lfs,
                                                                                             ),
                assignees=frozenset({name}),
                read_variables=frozenset({qp}),
                )


@symbol
def name_reference_gradient(element, restriction):
    name = restricted_name("js", restriction)
    evaluate_reference_gradient(element, name, restriction)
    return name


@cached
def evaluate_basis_gradient(element, name, restriction):
    lfs = name_lfs(element, restriction)
    temporary_variable(name, shape=(name_lfs_bound(lfs), name_dimension()), shape_impl=('vec', 'fv'))
    jac = name_jacobian_inverse_transposed(restriction)
    index = lfs_iname(element, restriction, context='transformgrads')
    reference_gradients = name_reference_gradient(element, restriction)
    instruction(inames=(index,
                        quadrature_iname(),
                        ),
                code='{}.mv({}[{}][0], {}[{}]);'.format(jac,
                                                        reference_gradients,
                                                        index,
                                                        name,
                                                        index,
                                                        ),
                assignees=name,
                read_variables=frozenset({reference_gradients}),
                )


@symbol
def name_basis_gradient(element, restriction):
    name = restricted_name("gradphi", restriction)
    evaluate_basis_gradient(element, name, restriction)
    return name


@cached
def evaluate_trialfunction(element, name, restriction):
    temporary_variable(name, shape=())
    lfs = name_lfs(element, restriction)
    index = lfs_iname(element, restriction, context='trial')
    basis = name_basis(element, restriction)
    from dune.perftool.pdelab.argument import pymbolic_coefficient
    coeff = pymbolic_coefficient(lfs, index, restriction)
    reduction_expr = Product((coeff, Subscript(Variable(basis), Variable(index))))
    instruction(expression=Reduction("sum", index, reduction_expr, allow_simultaneous=True),
                assignee=Variable(name),
                forced_iname_deps=frozenset({quadrature_iname()}),
                forced_iname_deps_is_final=True,
                )


@cached
def evaluate_trialfunction_gradient(element, name, restriction):
    temporary_variable(name, shape=(name_dimension(),), shape_impl=('fv',))
    lfs = name_lfs(element, restriction)
    index = lfs_iname(element, restriction, context='trialgrad')
    basis = name_basis_gradient(element, restriction)
    idim = dimension_iname()
    from dune.perftool.pdelab.argument import pymbolic_coefficient
    coeff = pymbolic_coefficient(lfs, index, restriction)
    reduction_expr = Product((coeff, Subscript(Variable(basis), (Variable(index), Variable(idim)))))
    instruction(expression=Reduction("sum", index, reduction_expr, allow_simultaneous=True),
                assignee=Subscript(Variable(name), Variable(idim)),
                forced_iname_deps=frozenset({quadrature_iname(), idim}),
                forced_iname_deps_is_final=True,
                )
